MODULE domvvl
   !!======================================================================
   !!                       ***  MODULE domvvl   ***
   !! Ocean : 
   !!======================================================================
   !! History :   9.0  !  06-06  (B. Levier, L. Marie)  original code
   !!              "   !  07-07  (D. Storkey) Bug fixes and code for BDY option.
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   'key_vvl'                              variable volume
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!   dom_vvl         : defined scale factors & depths at each time step
   !!   dom_vvl_ini     : defined coefficients to distribute ssh on each layers
   !!----------------------------------------------------------------------
   !! * Modules used
   USE oce             ! ocean dynamics and tracers
   USE dom_oce         ! ocean space and time domain
   USE sbc_oce         ! surface boundary condition: ocean
   USE dynspg_oce      ! surface pressure gradient variables
   USE phycst          ! physical constants
   USE in_out_manager  ! I/O manager
   USE lib_mpp         ! distributed memory computing library
   USE lbclnk          ! ocean lateral boundary conditions (or mpp link)
   USE bdy_oce         ! unstructured open boundary conditions

   IMPLICIT NONE
   PRIVATE

   !! * Routine accessibility
   PUBLIC dom_vvl_ini    ! called by dom_init.F90
   PUBLIC dom_vvl        ! called by istate.F90 and step.F90

   !! * Module variables
   REAL(wp), PUBLIC, DIMENSION(jpi,jpj) ::   hu_0, hv_0
   REAL(wp), PUBLIC, DIMENSION(jpi,jpj) ::   ee_t, ee_u, ee_v, ee_f   !:

   REAL(wp), PUBLIC, DIMENSION(jpi,jpj,jpk) ::   &  !:
      mut, muu, muv, muf                            !:

   REAL(wp), DIMENSION(jpk) ::   r2dt               ! vertical profile time-step, = 2 rdttra 
      !                                             ! except at nit000 (=rdttra) if neuler=0

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !!   OPA 9.0 , LOCEAN-IPSL (2005) 
   !! $Id$
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------

CONTAINS       

#if defined key_vvl

   SUBROUTINE dom_vvl_ini
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE dom_vvl_ini  ***
      !!                   
      !! ** Purpose :  compute coefficients muX at T-U-V-F points to spread
      !!               ssh over the whole water column (scale factors)
      !!
      !!----------------------------------------------------------------------
      INTEGER  ::   ji, jj, jk
      REAL(wp) ::   zcoefu, zcoefv, zcoeff
      !!----------------------------------------------------------------------

      IF(lwp)   THEN
         WRITE(numout,*)
         WRITE(numout,*) 'dom_vvl_ini : Variable volume activated'
         WRITE(numout,*) '~~~~~~~~~~~   compute coef. used to spread ssh over each layers'
      ENDIF

#if defined key_zco  ||  defined key_dynspg_rl
      CALL ctl_stop( 'dom_vvl_ini : options key_zco/key_dynspg_rl are incompatible with variable volume option key_vvl')
#endif

      fsdept(:,:,:) = gdept (:,:,:)
      fsdepw(:,:,:) = gdepw (:,:,:)
      fsde3w(:,:,:) = gdep3w(:,:,:)
      fse3t (:,:,:) = e3t   (:,:,:)
      fse3u (:,:,:) = e3u   (:,:,:)
      fse3v (:,:,:) = e3v   (:,:,:)
      fse3f (:,:,:) = e3f   (:,:,:)
      fse3w (:,:,:) = e3w   (:,:,:)
      fse3uw(:,:,:) = e3uw  (:,:,:)
      fse3vw(:,:,:) = e3vw  (:,:,:)

      ! mu computation
      ! --------------
      ! define ee_t, u, v and f as in sigma coordinate (ee_t = 1/ht, ...)
      ee_t(:,:) = fse3t_0(:,:,1)        ! Lower bound : thickness of the first model level
      ee_u(:,:) = fse3u_0(:,:,1)
      ee_v(:,:) = fse3v_0(:,:,1)
      ee_f(:,:) = fse3f_0(:,:,1)
      DO jk = 2, jpkm1                   ! Sum of the masked vertical scale factors
         ee_t(:,:) = ee_t(:,:) + fse3t_0(:,:,jk) * tmask(:,:,jk)
         ee_u(:,:) = ee_u(:,:) + fse3u_0(:,:,jk) * umask(:,:,jk)
         ee_v(:,:) = ee_v(:,:) + fse3v_0(:,:,jk) * vmask(:,:,jk)
         DO jj = 1, jpjm1                      ! f-point : fmask=shlat at coasts, use the product of umask
            ee_f(:,jj) = ee_f(:,jj) + fse3f_0(:,jj,jk) *  umask(:,jj,jk) * umask(:,jj+1,jk)
         END DO
      END DO  
      !                                  ! Compute and mask the inverse of the local depth at T, U, V and F points
      ee_t(:,:) = 1. / ee_t(:,:) * tmask(:,:,1)
      ee_u(:,:) = 1. / ee_u(:,:) * umask(:,:,1)
      ee_v(:,:) = 1. / ee_v(:,:) * vmask(:,:,1)
      DO jj = 1, jpjm1                         ! f-point case fmask cannot be used 
         ee_f(:,jj) = 1. / ee_f(:,jj) * umask(:,jj,1) * umask(:,jj+1,1)
      END DO
      CALL lbc_lnk( ee_f, 'F', 1. )       ! lateral boundary condition on ee_f
      !
      DO jk = 1, jpk
         mut(:,:,jk) = ee_t(:,:) * tmask(:,:,jk)   ! at T levels
         muu(:,:,jk) = ee_u(:,:) * umask(:,:,jk)   ! at T levels
         muv(:,:,jk) = ee_v(:,:) * vmask(:,:,jk)   ! at T levels
      END DO
      DO jk = 1, jpk
         DO jj = 1, jpjm1                      ! f-point : fmask=shlat at coasts, use the product of umask
               muf(:,jj,jk) = ee_f(:,jj) * umask(:,jj,jk) * umask(:,jj+1,jk)   ! at T levels
         END DO
         muf(:,jpj,jk) = 0.e0
      END DO
      CALL lbc_lnk( muf, 'F', 1. )       ! lateral boundary condition on ee_f


!!debug print
!         ii=50   ;   ij = 50
!         do jk= 1, jpk 
!         WRITE(numout,*) 'domvvl GM  : h0=', SUM( fse3t_0(ii,ij,1:jk) * tmask(ii,ij,1:jk) ), 'e3t0=', fse3t_0(ii,ij,jk),   &
!            &            'e3t =', fse3t_0(ii,ij,jk) * ( 1 +  mut(ii,ij,jk) ), 'mut', mut(ii,ij,jk),   &
!            &            'h =', SUM( fse3t_0(ii,ij,1:jk) * ( 1 +  mut(ii,ij,1:jk) ) * tmask(ii,ij,1:jk) )
!         end do
!!end debug print

      ! Reference ocean depth at U- and V-points
      hu_0(:,:) = 0.e0    
      hv_0(:,:) = 0.e0
      DO jk = 1, jpk
         hu_0(:,:) = hu_0(:,:) + fse3u(:,:,jk) * umask(:,:,jk)
         hv_0(:,:) = hv_0(:,:) + fse3v(:,:,jk) * vmask(:,:,jk)
      END DO

      ! before and now Sea Surface Height at u-, v-, f-points
      DO jj = 1, jpjm1
         DO ji = 1, jpim1
            zcoefu = 0.5  * umask(ji,jj,1) / ( e1u(ji,jj) * e2u(ji,jj) )
            zcoefv = 0.5  * vmask(ji,jj,1) / ( e1v(ji,jj) * e2v(ji,jj) )
            zcoeff = 0.25 * fmask(ji,jj,1)
!!gm bug used of fmask, even if thereafter multiplied by muf  which is correctly masked)
            sshu_b(ji,jj) = zcoefu * ( e1t(ji  ,jj) * e2t(ji  ,jj) * sshb(ji  ,jj)     &
               &                     + e1t(ji+1,jj) * e2t(ji+1,jj) * sshb(ji+1,jj) )   
            sshv_b(ji,jj) = zcoefv * ( e1t(ji,jj  ) * e2t(ji,jj  ) * sshb(ji,jj  )     &
               &                     + e1t(ji,jj+1) * e2t(ji,jj+1) * sshb(ji,jj+1) )   
            sshf_b(ji,jj) = zcoeff * ( sshb(ji  ,jj) + sshb(ji  ,jj+1)                 &
               &                     + sshb(ji+1,jj) + sshb(ji+1,jj+1) )               
            sshu_n(ji,jj) = zcoefu * ( e1t(ji  ,jj) * e2t(ji  ,jj) * sshn(ji  ,jj)     &
               &                     + e1t(ji+1,jj) * e2t(ji+1,jj) * sshn(ji+1,jj) )   
            sshv_n(ji,jj) = zcoefv * ( e1t(ji,jj  ) * e2t(ji,jj  ) * sshn(ji,jj  )     & 
               &                     + e1t(ji,jj+1) * e2t(ji,jj+1) * sshn(ji,jj+1) )     
            sshf_n(ji,jj) = zcoeff * ( sshn(ji  ,jj) + sshn(ji  ,jj+1)                 &
               &                     + sshn(ji+1,jj) + sshn(ji+1,jj+1) )               
         END DO
      END DO
      ! Boundaries conditions
      CALL lbc_lnk( sshu_b, 'U', 1. )   ;   CALL lbc_lnk( sshu_n, 'U', 1. )
      CALL lbc_lnk( sshv_b, 'V', 1. )   ;   CALL lbc_lnk( sshv_n, 'V', 1. )
      CALL lbc_lnk( sshf_b, 'F', 1. )   ;   CALL lbc_lnk( sshf_n, 'F', 1. )
      !


   END SUBROUTINE dom_vvl_ini


   SUBROUTINE dom_vvl
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE dom_vvl  ***
      !!                   
      !! ** Purpose :  compute ssh at U-V-F points, T-W scale factors and local
      !!               depths at each time step.
      !!----------------------------------------------------------------------
      !! * Local declarations
      INTEGER  ::   ji, jj, jk                 ! dummy loop indices
      !!----------------------------------------------------------------------

 !     IF( kt == nit000 ) THEN
 !        IF(lwp) WRITE(numout,*)
 !        IF(lwp) WRITE(numout,*) 'dom_vvl : '
 !        IF(lwp) WRITE(numout,*) '~~~~~~~ '
 !     ENDIF

      ! Scale factors at T levels
      DO jk = 1, jpkm1
         fse3t (:,:,jk) = fse3t_n (:,:,jk)
         fse3u (:,:,jk) = fse3u_n (:,:,jk)
         fse3v (:,:,jk) = fse3v_n (:,:,jk)
         fse3f (:,:,jk) = fse3f_n (:,:,jk)
         fse3w (:,:,jk) = fse3w_n (:,:,jk)
         fse3uw(:,:,jk) = fse3uw_n(:,:,jk)
         fse3vw(:,:,jk) = fse3vw_n(:,:,jk)
      END DO

      DO jk = 1, jpkm1
         DO jj = 1, jpj
            DO ji = 1, jpi
               fsdept(ji,jj,jk) = fsdept_n(ji,jj,jk)
               fsdepw(ji,jj,jk) = fsdepw_n(ji,jj,jk)
               fsde3w(ji,jj,jk) = fsde3w_n(ji,jj,jk)
            END DO
         END DO
      END DO

      ! Local depth or Inverse of the local depth of the water column at u- and v-points
      ! ------------------------------
      hu(:,:) = hu_0(:,:) + sshu_n(:,:)
      hv(:,:) = hv_0(:,:) + sshv_n(:,:)

      ! masked  inverse of the local depth
      hur(:,:) = 1. / MAX( hu(:,:), fse3u_0(:,:,1) ) * umask(:,:,1)
      hvr(:,:) = 1. / MAX( hv(:,:), fse3v_0(:,:,1) ) * vmask(:,:,1)

   END SUBROUTINE dom_vvl

#else
   !!----------------------------------------------------------------------
   !!   Default option :                                      Empty routine
   !!----------------------------------------------------------------------
   SUBROUTINE dom_vvl_ini
   END SUBROUTINE dom_vvl_ini
   SUBROUTINE dom_vvl
   END SUBROUTINE dom_vvl
#endif

   !!======================================================================
END MODULE domvvl
